import pytest
import numpy as np
import changepoynt.algorithms.sst as ssts
import logging
logging.basicConfig(level=logging.DEBUG)
LOGGER = logging.getLogger(__name__)


class TestSST:
    def setup_method(self):
        # set a random seed
        np.random.seed(3455)

        # create a random steps signal of a certain length
        self.signal_length = 300
        x0 = 1 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 1
        x1 = 3 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 2
        x2 = 5 * np.ones(self.signal_length) + np.random.rand(self.signal_length) * 1.5
        x = np.hstack([x0, x1, x2])
        x += np.random.rand(x.size)
        self.signal = x

    def teardown_method(self):
        pass

    def test_all_methods(self):
        # initialize random default method
        sst = ssts.SingularSpectrumTransformation(30)

        # get the different method names
        methods = list(sst.methods.keys())

        # go through the methods and check execution
        for method in methods:
            LOGGER.info(f'Starting SST for method {method}')
            ssts.SingularSpectrumTransformation(min(5, self.signal_length//2), rank=2,
                                                method=method).transform(self.signal)

    def test_default(self):
        ssts.SingularSpectrumTransformation(min(5, self.signal_length//2), rank=2).transform(self.signal)

    def test_unknown_method(self):
        with pytest.raises(AssertionError):
            ssts.SingularSpectrumTransformation(10, method='asdafwegrhqh')


if __name__ == "__main__":
    pytest.main()



# this file contains utility functions for handling and computing matrices needed for some of the algorithms in this
# package
import numpy as np
from numba import jit
from scipy.linalg import eigh_tridiagonal
from scipy.sparse.linalg import svds, eigsh
import fbpca
import warnings


@jit(nopython=True)
def power_method(a_matrix: np.ndarray, x_vector: np.ndarray, n_iterations: int) -> (float, np.ndarray):
    """
    This function searches the largest (dominant) eigenvalue and corresponding eigenvector by repeated multiplication
    of the matrix A with an initial vector. It assumes a dominant eigenvalue bigger than the second one, otherwise
    it won't converge.

    For proof and explanation look at:
    https://pythonnumericalmethods.berkeley.edu/notebooks/chapter15.02-The-Power-Method.html
    :param a_matrix: 2D-Matrix of size NxN filled with floats
    :param x_vector: Vector of size Nx1 filled with floats
    :param n_iterations: the amount of iterations for the approximation
    :return: the dominant eigenvalue and corresponding eigenvector
    """

    # go through the iterations and continue to scale the returned vector, so we do not reach extreme values
    # during the iteration we scale the vector by its maximum as we can than easily extract the eigenvalue
    a_square = a_matrix.T @ a_matrix
    for _ in range(n_iterations):

        # multiplication with a_matrix.T @ a_matrix as can be seen in explanation of
        # https://numpy.org/doc/stable/reference/generated/numpy.linalg.svd.html
        x_vector = a_square @ x_vector

        # scale the vector so we keep the values in bound
        x_vector = x_vector / np.max(x_vector)

    # get the normed eigenvector
    x_vector = x_vector / np.linalg.norm(x_vector)

    # get the corresponding eigenvalue
    eigenvalue = np.linalg.norm(a_matrix @ x_vector)
    return eigenvalue, (a_matrix @ x_vector)/eigenvalue


@jit(nopython=True)
def lanczos(a_matrix: np.ndarray, r_0: np.ndarray, k: int) -> (np.ndarray, np.ndarray):
    """
    This function computes the tri-diagonalization matrix from the square matrix C which is the result of the lanczos
    algorithm.

    The algorithm has been described and proven in:
    Idé, Tsuyoshi, and Koji Tsuda.
    "Change-point detection using krylov subspace learning."
    Proceedings of the 2007 SIAM International Conference on Data Mining.
    Society for Industrial and Applied Mathematics, 2007.

    :param a_matrix: 2D-Matrix of size NxN filled with floats where we want to find the krylov subspace approx. for
    :param r_0: intial starting vector for the subspace approximation
    :param k: size of the approximation
    :return: Returns the alpha and beta values for the tridiagonal, symmetric matrix T. alphas are the values from the
    main diagonal and beta from the off diagonal as described in
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.eigh_tridiagonal.html (alpha = d, beta = e)
    """

    # save the initial vector
    r_i = r_0
    q_i = np.zeros_like(r_i)

    # initialization of the diagonal elements
    alphas = np.zeros(shape=(k + 1,), dtype=np.float64)
    betas = np.ones(shape=(k + 1,), dtype=np.float64)

    # Subroutine 1 of the paper
    for j in range(k + 1):
        # compute r_(j+1)
        new_q = r_i / betas[j]

        # compute the new alpha
        alphas[j + 1] = new_q.T @ a_matrix @ new_q

        # compute the new r
        r_i = a_matrix @ new_q - alphas[j + 1] * new_q - betas[j] * q_i

        # compute the next beta
        betas[j + 1] = np.linalg.norm(r_i)

        # update the previous q
        q_i = new_q

    return alphas[1:], betas[1:-1]


def tridiagonal_eigenvalues(alphas: np.ndarray, betas: np.ndarray, amount=-1):
    """
    This function uses a fast approach for symmetric tridiagonal matrices to calculate the [amount] highest eigenvalues
    and corresponding eigenvectors.

    :param alphas: main diagonal elements
    :param betas: off diagonal elements
    :param amount: The amount of eigenvalues you want to compute (from the highest)
    :return: eigenvalues and corresponding eigenvectors
    """

    # check whether we need to use default parameters
    if amount < 0:
        amount = alphas.shape[0]

    # assertions about shape and dimensions as well as amount of eigenvectors
    assert 0 < amount <= alphas.shape[0], 'We can only calculate one to size of matrix eigenvalues.'
    assert alphas.ndim == 1, 'The alphas need to be vectors.'
    assert betas.ndim == 1, 'The betas need to be vectors.'
    assert alphas.shape[0] - 1 == betas.shape[0], 'Alpha size needs to be exactly one bigger than beta size.'

    # compute the decomposition
    eigenvalues, eigenvectors = eigh_tridiagonal(d=alphas, e=betas,
                                                 select='i', select_range=(alphas.shape[0]-amount, alphas.shape[0]-1))

    # return them to be in sinking order
    return eigenvalues[::-1], eigenvectors[:, ::-1]


def rayleigh_ritz_singular_value_decomposition(a_matrix: np.ndarray, k: int) -> (np.ndarray, np.ndarray):
    """
    This function uses the Rayleigh-Ritz method implemented in ARPACK to compute the k highest eigenvalues and
    corresponding eigenvectors. It should be faster than a complete svd.

    !NOTE!:
    The order of the k highest eigenvalues is not guaranteed by this method!

    :param a_matrix: 2D-Matrix filled with floats for which we want to find the left eigenvectors
    :param k: the amount of highest eigenvectors we want to find
    :return: returns the eigenvalues and eigenvectors as numpy arrays
    """
    eigenvectors, eigenvalues, _ = svds(a_matrix, k=k)
    return eigenvalues, eigenvectors


def randomized_singular_value_decomposition(a_matrix: np.ndarray, randomized_rank: int) -> (np.ndarray, np.ndarray):
    """
    This function implements randomized singular vector decomposition of a matrix as surveyed and described in

    Halko, Nathan, Per-Gunnar Martinsson, and Joel A. Tropp.
    "Finding structure with randomness: Probabilistic algorithms for constructing approximate matrix decompositions."
    SIAM review 53.2 (2011): 217-288.

    on page 4 chapter 1.3 and further.

    It also incoporates ideas from

    Szlam, Arthur, Yuval Kluger, and Mark Tygert.
    "An implementation of a randomized algorithm for principal component analysis."
    arXiv preprint arXiv:1412.3510 (2014).

    in order to further imprint the highest eigenvectors into the approximation matrix after multiplying with the
    randomized matrix.

    This implementation is generally inspired by
    https://scikit-learn.org/stable/modules/generated/sklearn.utils.extmath.randomized_svd.html
    but reimplemented as we wanted to save dependencies and use jit compiled code.

    :param a_matrix: 2D-Matrix filled with floats for which we want to find the left eigenvectors
    :param randomized_rank: the rank of the noise matrix used for randomized svd. the higher the rank, the better the
    approximation but the lower the precision of the eigenvectors
    :return:
    """

    # construct the random matrix for multiplication with the matrix we want to decompose
    approximation_matrix = np.random.normal(size=(a_matrix.shape[1], randomized_rank))

    # perform power iterations to further "enhance" the top singular of the hankel matrix into Q
    a_square = a_matrix.T @ a_matrix
    for _ in range(3):
        approximation_matrix = a_square @ approximation_matrix

    # extract the orthonormal base of the approximation matrix
    q_substitute, _ = np.linalg.qr(a_matrix @ approximation_matrix)

    # project the hankel matrix onto the lower dimensional orthonormal basis of the approximation matrix
    base_projection = q_substitute.T @ a_matrix

    # compute the singular vector decomposition of the thinner matrix base projection
    eigenvectors, eigenvalues, _ = np.linalg.svd(base_projection, full_matrices=False)

    # compute the original eigenvectors
    eigenvectors = np.dot(q_substitute, eigenvectors)

    return eigenvalues, eigenvectors


def facebook_randomized_svd(a_matrix: np.ndarray, randomized_rank: int) -> (np.ndarray, np.ndarray):
    eigenvectors, eigenvalues, _ = fbpca.pca(a_matrix, randomized_rank, True)
    return eigenvalues, eigenvectors


def implicit_restarted_lanczos_bidiagonalization(a_matrix: np.ndarray, rank: int,
                                                 lanczos_rank: int) -> (np.ndarray, np.ndarray):
    """
    This function uses the implicitly Restarted Lanczos method implemented in ARPACK to compute the k highest
    eigenvalues and corresponding eigenvectors. It should be faster as a complete svd.

    !NOTE!
    This method only works for symmetric (hermitian) matrices!!

    :param a_matrix: 2D-Matrix filled with floats for which we want to find the left eigenvectors
    :param rank: the amount of highest eigenvectors we want to find
    :param lanczos_rank: the size of the lanczos subspace approximation
    :return: returns the eigenvalues and eigenvectors as numpy arrays
    """
    eigenvalues, eigenvectors = eigsh(a_matrix, k=rank, which='LM', ncv=lanczos_rank)
    return eigenvalues, eigenvectors


def examples():
    """
    This function implements some usage examples for quick internal testing. It is not aimed for beeing used.
    :return: None
    """

    import time

    # set a random seed
    np.random.seed(1234)

    # create the random starting vector
    x = np.random.rand(100)
    x /= np.linalg.norm(x)
    A = np.random.rand(100, 100)

    # test the power method
    eigval, eigvec = power_method(A, x, n_iterations=100)
    eigvecs, eigvals, _ = np.linalg.svd(A)
    print(eigval, eigvals[0])

    # test the tridiagonalization method
    size = 1000
    d = 3 * np.random.rand(size)
    e = -1 * np.random.rand(size - 1)
    start = time.time()
    tri_eigvals, tri_eigvecs = tridiagonal_eigenvalues(d, e, size // 2)
    print(f'Specialized tridiagonal SVD took: {time.time() - start} s.')
    T = np.diag(d) + np.diag(e, k=1) + np.diag(e, k=-1)
    start = time.time()
    eigvecs, eigvals , _ = np.linalg.svd(T)
    print(f'Normal SVD took: {time.time() - start} s.')

    # test randomized svd
    eigval, eigvec = randomized_singular_value_decomposition(A, 20)
    eigvecs, eigvals, _ = np.linalg.svd(A)
    print(eigval, eigvals[0])

    # test irlbd
    eigval, eigvec, = implicit_restarted_lanczos_bidiagonalization(A, 20)
    eigvecs, eigvals, _ = np.linalg.svd(A)
    print(eigval, eigvals[0])


@jit(nopython=True)
def compile_hankel(time_series: np.ndarray, end_index: int, window_size: int, rank: int) -> np.ndarray:
    """
    This function constructs a hankel matrix from a 1D time series. Please make sure constructing the matrix with
    the given parameters (end index, window size, etc.) is possible, as this function does no checks due to
    performance reasons.

    :param time_series: 1D array with float values as the time series
    :param end_index: the index (point in time) where the time series starts
    :param window_size: the size of the windows cut from the time series
    :param rank: the amount of time series in the matrix
    :return: The hankel matrix with lag one
    """

    # make an empty matrix to place the values
    #
    # almost no faster way:
    # https://stackoverflow.com/questions/71410927/vectorized-way-to-construct-a-block-hankel-matrix-in-numpy-or-scipy
    hankel = np.empty((window_size, rank))

    # go through the time series and make the hankel matrix
    for cx in range(rank):
        hankel[:, cx] = time_series[(end_index-window_size-cx):(end_index-cx)]
    return hankel


if __name__ == '__main__':
    examples()



import numpy as np
from numba import jit
from changepoynt.utils import linalg as lg
from changepoynt.utils import normalization
from typing import Callable
from functools import partial
from changepoynt.algorithms.base_algorithm import Algorithm


class SingularSpectrumTransformation(Algorithm):
    """
    This class implements all the utility and functionality necessary to compute the SST change point detection
    algorithm as described in:

    [1]
    Idé, Tsuyoshi, and Keisuke Inoue.
    "Knowledge discovery from heterogeneous dynamic systems using change-point correlations."
    Proceedings of the 2005 SIAM international conference on data mining.
    Society for Industrial and Applied Mathematics, 2005.

    [2]
    Idé, Tsuyoshi, and Koji Tsuda.
    "Change-point detection using krylov subspace learning."
    Proceedings of the 2007 SIAM International Conference on Data Mining.
    Society for Industrial and Applied Mathematics, 2007.

    It also uses a technique called randomized singular value decomposition which is surveyed and described in

    [3]
    Halko, Nathan, Per-Gunnar Martinsson, and Joel A. Tropp.
    "Finding structure with randomness: Probabilistic algorithms for constructing approximate matrix decompositions."
    SIAM review 53.2 (2011): 217-288.

    There will be a parameter specifying, whether to use the implicit krylov approximation from [2]. This
    significantly speeds up the computation but can reduce accuracy as the "eigensignals" are only approximated
    indirectly using a krylov subspace with the possible change point signal as seed. This method also deploys
    a feedback of the dominant "eigensignal" as seed into the power method (dominant eigenvector estimation)
    with additive gaussian noise.

    !NOTE!:
    Most computational heavy functions are implemented as standalone functions, even if they require instance variables
    as this enables us to use jit compiled code provided by the numba compiler for faster calculations.

    Also, we decided to do type hinting but not type checking as it requires too much boilerplate code. We recommend
    to input the specified types as the program will break in unforeseen ways otherwise.
    """

    def __init__(self, window_length: int, n_windows: int = None, lag: int = None, rank: int = 5, scale: bool = True,
                 method: str = 'ika', lanczos_rank: int = None, random_rank: int = None,
                 feedback_noise_level: float = 1e-3) -> None:
        """
        Initializing the Singular Spectrum Transformation (SST) requires setting a lot of parameters. See the parameters
        explanation for some intuition into the right choices. Currently, there are two SST methods from [1] and [2]
        available for use.

        :param window_length: This specifies the length of the time series (in samples), which will be used to extract
        the representative "eigensignals" to be compared before and after the lag. The windows length should be big
        enough to cover any wanted patterns (e.g. bigger than the periodicity of periodic signals).

        :param n_windows: This specifies the amount of consecutive time windows used to extract the "eigensignals" from
        the given time series. It should be big enough to cover different parts of the target behavior. If one does not
        specify this value, we use the rule of thumb and take as many time windows as you specified the length of the
        window (rule of thumb).

        :param lag: This parameter specifies the distance of the comparison for behaviors. In easy terms it tells the
        algorithms how far it should look into the future to find change in behavior to the current signal. If you do
        not specify this parameter, we use a rule of thumb and look ahead half of the window length you specified to
        cover the behavior.

        :param rank: This parameter specifies the amount of "eigensignals" which will be used to measure the
        dissimilarity of the signal in the future behavior. As a rule of thumb we take the five most dominant
        "eigensignals" if you do no specify otherwise.

        :param scale: Due to numeric stability we REALLY RECOMMEND scaling the signal into a restricted value range. Per
        default, we use a min max scaling to ensure a restricted value range. In the presence of extreme outliers this
        could cause problems, as the signal will be squished.

        :param method: Currently "svd" [1] and "ika" [2] are available. ika corresponds to IKA-SST in [2] which
        will speed up the computation significantly.

        :param lanczos_rank: In order to use the implicit approximation of "eigensignals" by using "ika" [2] method
        one needs to decide the rank of the implicit approximation of each "eigensignal". As a rule of thumb we
        determine this value as beeing twice the amount of specified "eigensignals" to span the subspace
        (parameter rank). This is also recommend by the author of IKA-SST in [2].

        :param random_rank: In order to use the randomized singular value decomposition, one needs to provide a
        randomized rank (size of second matrix dimension for the randomized matrix) as specified in [3]. The lower
        this value, the faster the computation but the higher the error (as the approximation gets worse).

        :param feedback_noise_level: This specifies the amplitude of additive white gaussian noise added to the dominant
        "eigensignal" of the future behavior when shifting forward. This idea is noted in [2] and initializes
        the seed of the power method for dominant eigenvector estimation with the precious dominant eigenvector
        plus the noise level specified here. The noise level should just be a small fraction of the value range
        of the signal.
        """

        # save the specified parameters into instance variables
        self.window_length = window_length
        self.n_windows = n_windows
        self.lag = lag
        self.rank = rank
        self.scale = scale
        self.method = method
        self.lanczos_rank = lanczos_rank
        self.random_rank = random_rank
        self.noise = feedback_noise_level

        # set some default values when they have not been specified
        if self.n_windows is None:
            self.n_windows = self.window_length
        if self.lag is None:
            # rule of thumb
            self.lag = max(self.n_windows//3, 1)
        if self.lanczos_rank is None:
            # make rank even and multiply by two just as specified in [2]
            self.lanczos_rank = self.rank * 2 - (self.rank & 1)
        if self.random_rank is None:
            # compute the rank as specified in [3] and
            # https://scikit-learn.org/stable/modules/generated/sklearn.utils.extmath.randomized_svd.html
            self.random_rank = min(self.rank + 10, self.window_length, self.n_windows)

        # specify the methods and their corresponding functions as lambda functions expecting only the
        # 1) future hankel matrix,
        # 2) the current hankel matrix and
        # 3) the feedback vector (e.g. for dominant eigenvector feedback)
        # all other parameter should be specified as values in the partial lambda function
        self.methods = {'ika': partial(_implicit_krylov_approximation,
                                       rank=self.rank,
                                       lanczos_rank=self.lanczos_rank),
                        'svd': partial(_rayleigh_singular_value_decomposition,
                                       rank=self.rank),
                        'rsvd': partial(_randomized_singular_value_decomposition,
                                        rank=self.rank,
                                        randomized_rank=self.random_rank),
                        'fbrsvd': partial(_facebook_random_singular_value_decomposition,
                                          rank=self.rank,
                                          randomized_rank=self.random_rank)
                        }

        # check whether the method is correct
        assert self.method in self.methods, f'Specified method {self.method} is not available in {self.methods.keys()}.'

    def transform(self, time_series: np.ndarray) -> np.ndarray:
        """
        This function calculate the anomaly score for each sample within the time series starting from an initial sample
        being the first sample of fitting the past hankel matrix (window_length + n_windows samples) and the new future
        hankel matrix (+lag). Values before that will be zero

        It also does some assertions regarding the specified parameters and whether they fit the time series.

        This function builds the interface to the jit compiled static function used to iterate over the array.

        :param time_series: 1D array containing the time series to be scored
        :return: anomaly score
        """

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # compute the starting point of the scoring (past and future hankel need to fit)
        starting_point = self.window_length + self.n_windows + self.lag
        assert starting_point < time_series.shape[0], "The time series is too short to score any points."

        # scale the time series (or just copy it if already scaled)
        if self.scale:
            time_series = normalization.min_max_scaling(time_series, min_val=1.0, max_val=2.0, inplace=False)
        else:
            time_series = time_series.copy()

        # get the changepoint scorer from the different methods
        scoring_function = self.methods[self.method]

        # start the scaling itself by calling the jit compiled staticmethod and return the result
        score = _transform(time_series=time_series, start_idx=starting_point, window_length=self.window_length,
                           n_windows=self.n_windows, lag=self.lag, scoring_function=scoring_function)
        return score


def _transform(time_series: np.ndarray, start_idx: int, window_length: int, n_windows: int, lag: int,
               scoring_function: Callable) -> np.ndarray:
    """
    Compute heavy and hopefully jit compilable score computation for the SST method. It does not do any parameter
    checking and can throw cryptic errors. It's only used for internal use as a private function.

    :param time_series: 1D array containing the time series to be scored
    :param start_idx: integer defining the start sample index for the score computation
    :param window_length: the size of the time series window for each column of the hankel matrix
    :param n_windows: amount of columns in the hankel matrix
    :param lag: sample distance between future and past hankel matrix
    """

    # create initial vector for ika method with feedback dominant eigenvector as proposed in [2]
    # with a norm of one
    x0 = np.random.rand(n_windows)
    x0 /= np.linalg.norm(x0)

    # initialize a scoring array with no values yet
    score = np.zeros_like(time_series)

    # iterate over all the values in the signal starting at start_idx computing the change point score
    for idx in range(start_idx, time_series.shape[0]):

        # compile the past hankel matrix (H1)
        hankel_past = lg.compile_hankel(time_series, idx-lag, window_length, n_windows)

        # compile the future hankel matrix (H2)
        hankel_future = lg.compile_hankel(time_series, idx, window_length, n_windows)

        # compute the score and save the returned feedback vector
        score[idx], x1 = scoring_function(hankel_past, hankel_future, x0)

        # add noise to the dominant eigenvector and normalize it again
        x0 = x1 + 1e-3 * np.random.rand(x0.size)
        x0 /= np.linalg.norm(x0)

    return score


def _implicit_krylov_approximation(hankel_past: np.ndarray, hankel_future: np.ndarray, x0: np.ndarray,
                                   rank: int, lanczos_rank: int) -> (float, np.ndarray):
    """
    This function computes the change point score based on the krylov subspace approximation of the SST as proposed in

    Idé, Tsuyoshi, and Koji Tsuda.
    "Change-point detection using krylov subspace learning."
    Proceedings of the 2007 SIAM International Conference on Data Mining.
    Society for Industrial and Applied Mathematics, 2007.

    Due to perfomance reason this function makes no sanity checks for matrix size and parameter validity.

    :param hankel_past: the hankel matrix H1 before the change point
    :param hankel_future: the hankel matrix H2 after the change point
    :param x0: the initialization value for the power method applied to H2 to find the dominant eigenvector
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :param lanczos_rank: the rank of the approximation of the "eigensignals"
    :return: the change point score, the new dominant eigenvector of H2 for the feedback into the next H2
    """

    # compute the biggest eigenvector of the hankel matrix after the possible change point (h2)
    c_2 = hankel_future.T @ hankel_future
    _, eigvec_future = lg.power_method(c_2, x0, n_iterations=1)

    # compute the empirical covariance matrix before the possible change point (H1)
    c_1 = hankel_past.T @ hankel_past

    # compute the tridiagonal matrix from c1
    alphas, betas = lg.lanczos(c_1, eigvec_future, lanczos_rank)

    # compute the singular value decomposition of the tridiagonal matrix (only the biggest)
    _, eigvecs = lg.tridiagonal_eigenvalues(alphas, betas, rank)

    # compute the similarity score as defined in the ika sst paper and also return our u for the
    # feedback loop in figure 3 of the paper
    return 1 - (eigvecs[0, :] * eigvecs[0, :]).sum(), eigvec_future


def _rayleigh_singular_value_decomposition(hankel_past: np.ndarray, hankel_future: np.ndarray, x0: np.ndarray,
                                           rank: int) -> (float, np.ndarray):
    """
    This function implements change point detection using rayleigh-ritz singular value decomposition
    and computes the change point score as proposed in:

    Idé, Tsuyoshi, and Keisuke Inoue.
    "Knowledge discovery from heterogeneous dynamic systems using change-point correlations."
    Proceedings of the 2005 SIAM international conference on data mining.
    Society for Industrial and Applied Mathematics, 2005.

    :param hankel_past: the hankel matrix H1 before the change point
    :param hankel_future: the hankel matrix H2 after the change point
    :param x0: highest eigenvector of previous iteration (will be ignored in this function and is just added to complete
    the function signature, i.e. input and output size, to be compatible with other methods)
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :return: the change point score, the input vector x0
    """

    # compute the rank highest eigenvectors of the past hankel matrix
    _, eigvecs_past = lg.rayleigh_ritz_singular_value_decomposition(hankel_past, rank)

    # compute the dominant eigenvector of the future time series
    c_2 = hankel_future.T @ hankel_future
    _, eigvec_future = lg.power_method(c_2, x0, n_iterations=1)

    # compute the projection distance
    alpha = eigvecs_past.T @ eigvec_future
    return 1 - alpha.T @ alpha, x0


def _lanczos_singular_value_decomposition(hankel_past: np.ndarray, hankel_future: np.ndarray, x0: np.ndarray,
                                          rank: int, lanczos_rank: int) -> (float, np.ndarray):
    """
    This function implements change point detection using restarted implicit lanczos singular value decomposition
    and computes the change point score as proposed in:

    Idé, Tsuyoshi, and Keisuke Inoue.
    "Knowledge discovery from heterogeneous dynamic systems using change-point correlations."
    Proceedings of the 2005 SIAM international conference on data mining.
    Society for Industrial and Applied Mathematics, 2005.

    !NOTE!
    Does currently not converge and therefore is not in use.

    :param hankel_past: the hankel matrix H1 before the change point
    :param hankel_future: the hankel matrix H2 after the change point
    :param x0: highest eigenvector of previous iteration (will be ignored in this function and is just added to complete
    the function signature, i.e. input and output size, to be compatible with other methods)
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :param lanczos_rank: the rank of the approximation of the "eigensignals"
    :return: the change point score, the input vector x0
    """

    # compute the rank highest eigenvectors of the past hankel matrix
    _, eigvecs_past = lg.implicit_restarted_lanczos_bidiagonalization(hankel_past, rank, lanczos_rank=lanczos_rank)

    # compute the dominant eigenvector of the future time series
    c_2 = hankel_future.T @ hankel_future
    _, eigvec_future = lg.power_method(c_2, x0, n_iterations=1)

    # compute the projection distance
    alpha = eigvecs_past.T @ eigvec_future
    return 1 - alpha.T @ alpha, x0


def _randomized_singular_value_decomposition(hankel_past: np.ndarray, hankel_future: np.ndarray, x0: np.ndarray,
                                             rank: int, randomized_rank: int):
    """
    This function implements the idea of

    Idé, Tsuyoshi, and Keisuke Inoue.
    "Knowledge discovery from heterogeneous dynamic systems using change-point correlations."
    Proceedings of the 2005 SIAM international conference on data mining.
    Society for Industrial and Applied Mathematics, 2005.

    but uses a feedback vector for the future dominant eigenvector as proposed in

    Idé, Tsuyoshi, and Koji Tsuda.
    "Change-point detection using krylov subspace learning."
    Proceedings of the 2007 SIAM International Conference on Data Mining.
    Society for Industrial and Applied Mathematics, 2007.

    in contrast to the first two papers, the method uses randomized singular value decomposition.

    :param hankel_past: the hankel matrix H1 before the change point
    :param hankel_future: the hankel matrix H2 after the change point
    :param x0: the initialization value for the power method applied to H2 to find the dominant eigenvector
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :param randomized_rank: the rank of the approximation used to construct the noise matrix
    :return: the change point score, the new dominant eigenvector of H2 for the feedback into the next H2
    """

    # compute the biggest eigenvector of the hankel matrix after the possible change point (h2)
    c_2 = hankel_future.T @ hankel_future
    _, eigvec_future = lg.power_method(c_2, x0, n_iterations=1)

    # compute the eigenvectors of the past hankel matrix
    c_1 = hankel_past.T @ hankel_past
    _, eigenvecs_past = lg.randomized_singular_value_decomposition(c_1, randomized_rank=randomized_rank)

    # compute the change point score as defined in the papers
    alpha = eigenvecs_past[:, :rank].T @ eigvec_future
    return 1-alpha.T @ alpha, eigvec_future


def _facebook_random_singular_value_decomposition(hankel_past: np.ndarray, hankel_future: np.ndarray, x0: np.ndarray,
                                                  rank: int, randomized_rank: int):
    """
    This function implements the idea of

    Idé, Tsuyoshi, and Keisuke Inoue.
    "Knowledge discovery from heterogeneous dynamic systems using change-point correlations."
    Proceedings of the 2005 SIAM international conference on data mining.
    Society for Industrial and Applied Mathematics, 2005.

    but uses the randomized svd decomposition by

    Szlam, Arthur, Yuval Kluger, and Mark Tygert.
    "An implementation of a randomized algorithm for principal component analysis."
    arXiv preprint arXiv:1412.3510 (2014).

    :param hankel_past: the hankel matrix H1 before the change point
    :param hankel_future: the hankel matrix H2 after the change point
    :param x0: the initialization value for the power method applied to H2 to find the dominant eigenvector
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :param randomized_rank: the rank of the approximation used to construct the noise matrix
    :return: the change point score, the new dominant eigenvector of H2 for the feedback into the next H2
    """
    # compute the biggest eigenvector of the hankel matrix after the possible change point (h2)
    c_2 = hankel_future.T @ hankel_future
    _, eigvec_future = lg.power_method(c_2, x0, n_iterations=1)

    # compute the eigenvectors of the past hankel matrix
    c_1 = hankel_past.T @ hankel_past
    _, eigenvecs_past = lg.facebook_randomized_svd(c_1, randomized_rank=randomized_rank)

    # compute the change point score as defined in the papers
    alpha = eigenvecs_past[:, :rank].T @ eigvec_future
    return 1-alpha.T @ alpha, eigvec_future


if __name__ == '__main__':
    from time import time
    # make synthetic step function
    np.random.seed(123)
    length = 300
    x = np.hstack([1 * np.ones(length) + np.random.rand(length) * 1,
                   3 * np.ones(length) + np.random.rand(length) * 2,
                   5 * np.ones(length) + np.random.rand(length) * 1.5])
    x += np.random.rand(x.size)

    # create the sst method
    ika_sst = SingularSpectrumTransformation(31, method='ika')
    svd_sst = SingularSpectrumTransformation(31, method='svd')
    rsvd_sst = SingularSpectrumTransformation(31, method='rsvd')
    fbrsvd_sst = SingularSpectrumTransformation(31, method='fbrsvd')

    # make the scoring
    start = time()
    ika_sst.transform(x)
    print((time() - start) / (length * 3))
    svd_sst.transform(x)
    rsvd_sst.transform(x)
    fbrsvd_sst.transform(x)



# this file contains utility functions to scale and normalize signals
import numpy as np


def min_max_scaling(time_series: np.ndarray, min_val: float = 0.0, max_val: float = 1.0, inplace: bool = False)\
        -> np.ndarray:
    """
    This function applied min max scaling for an 1D-array. It is inspired by:
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html
    but lighter and reimplemented in order to not introduce unnecessary dependencies for small functionality.

    Min max scaling as implemented here is sensitive to extreme outliers, but guarantees the value range not
    exceeding min_val and max_val.

    :param time_series: 1D array containing consecutive values for one feature
    :param min_val: the minimum value the scaled time series will reach
    :param max_val: the maximum value the scale time series will reach
    :param inplace: boolean to specify whether the input array will be scaled and changed in place

    :return: the scaled input array.
    """
    # make some assertion checks
    assert time_series.ndim == 1, 'Time series needs to be an 1D array.'

    # copy the time series if specified
    if not inplace: time_series = time_series.copy()

    # scale the time series to values between zero and one
    time_series = (time_series-np.min(time_series, axis=0)) / (np.max(time_series, axis=0)-np.min(time_series, axis=0))

    # scale into the wished value range
    time_series = time_series * (max_val - min_val) + min_val
    return time_series



import numpy as np
from changepoynt.utils import linalg as lg
from changepoynt.utils import densityratioestimation as dre
from changepoynt.algorithms.base_algorithm import Algorithm


class RuLSIF(Algorithm):
    """
    This class implements change point detection based on relative density ration estimation optimizing a least squares
    approach from:

    [1]
    Liu, Song, et al.
    "Change-point detection in time-series data by relative density-ratio estimation."
    Neural Networks 43 (2013): 72-83.

    We use an implemented package called densratio for the computations of the density ratio and the Pearson divergence
    calculations. We merely adopt some steps to account for the symmetric Pearson divergence used in [1]. We adopted
    the function RuLSIF(x, y, alpha, sigma_range, lambda_range, kernel_num=100, verbose=True) from the pacakge.

    The code has been created looking at:
    - https://github.com/hoxo-m/densratio_py
    - https://pypi.org/project/densratio/
    - http://www.makotoyamada-ml.com/RuLSIF.html (Matlab implementation original author)

    and was adopted from:
    - https://github.com/DeepOmics/rulsif_abrupt-change_detection
    and we can not thank the authors enough for writing such deep mathematical code!
    The reason for adoption were the one named above (symmetric pearson divergence)
    and more direct use of numba as a jit compiler, as well as we do not want and use many things computed in the
    package, so we can save time by not computing them.
    """

    def __init__(self, window_length: int = 10, n_windows: int = 50, lag: int = None, estimation_lag: int = None,
                 n_kernels: int = 100, alpha: float = 0.1) -> None:
        """
        This defines all necessary parameters for the RuLSIF to work.
        :param window_length: the length of the windows we want to compare the densities for (k in the paper)
        :param n_windows: the amount of windows (n in the paper)
        :param lag: the difference in time between the past and the future comparison window
        :param estimation_lag: how often we should re-estimate lambda and sigma for the gaussian kernels
        :param alpha: the smoothing parameter for the RELATIVE in RuLSIF
        :param n_kernels: the amount of gaussian kernels for the density ratio
        """
        self.window_length = window_length
        self.n_windows = n_windows
        self.lag = lag
        self.estimation_lag = estimation_lag
        self.n_kernels = n_kernels
        self.alpha = alpha

        # check that alpha does not exceed the bounds
        assert 0 <= self.alpha < 1, 'The alpha parameter should be in the interval [0,1).'

        # check for the estimation lag
        assert self.estimation_lag is None or 1 <= self.estimation_lag, \
            'The estimation lag needs to be bigger than zero samples.'

        # set the lag if it is not given
        if not self.lag:
            self.lag = self.n_windows

        # create the estimator from the utils
        self.estimator = dre.RULSIF(alpha=self.alpha, n_kernels=self.n_kernels)

    def transform(self, time_series: np.ndarray):

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # check whether we want no updated estimation
        if self.estimation_lag is None:
            estimation_lag = len(time_series)+1
        else:
            estimation_lag = self.estimation_lag

        # compute the starting point of the scoring (past and future hankel need to fit)
        starting_point = self.window_length + self.n_windows + self.lag
        assert starting_point < time_series.shape[0], "The time series is too short to score any points."

        # copy the time series to protect it from any modifications within the algorithm
        time_series = time_series.copy()

        # call the function to compute the values
        return _transform(time_series, starting_point, self.window_length, self.n_windows, self.lag,
                          estimation_lag, self.estimator)


def _transform(time_series: np.ndarray, starting_point: int, window_length: int, n_windows: int, lag: int,
               estimation_lag: int, estimator: dre.Estimator) -> np.ndarray:

    # create the empty score vector
    score = np.zeros_like(time_series)

    # iterate over all the values in the signal starting at start_idx computing the change point score
    for idx in range(starting_point, time_series.shape[0]):

        # compile the past hankel matrix (Y)
        hankel_past = lg.compile_hankel(time_series, idx - lag, window_length, n_windows)

        # compile the future hankel matrix (Y')
        hankel_future = lg.compile_hankel(time_series, idx, window_length, n_windows)

        # check whether we need the cross validation for the kernel width and position
        if (idx-starting_point) % estimation_lag == 0:
            estimator.train(hankel_past, hankel_future)

        # compute the score and save the returned feedback vector
        score[idx] = estimator.apply(hankel_past, hankel_future) + estimator.apply(hankel_future, hankel_past)

    return score


def short_test():
    from time import time
    # make synthetic step function
    np.random.seed(123)
    length = 300
    x = np.hstack([1 * np.ones(length) + np.random.rand(length) * 1,
                   3 * np.ones(length) + np.random.rand(length) * 2,
                   5 * np.ones(length) + np.random.rand(length) * 1.5])
    x += np.random.rand(x.size)

    # create the rulsif method
    ruli = RuLSIF()

    # make the scoring
    start = time()
    score = ruli.transform(x)
    print((time() - start) / (length * 3))
    print(score[600:610])


if __name__ == '__main__':
    short_test()



import numpy as np
import scipy as sp
from changepoynt.utils import linalg as lg
from changepoynt.utils import normalization
from typing import Callable
from functools import partial


class SubspaceIdentification:
    """
    This class implements all the utility and functionality necessary to compute the subspace change point detection
    algorithm as described in (offline method)

    [1]
    "Change-Point Detection in Time-Series Data based on Subspace Identification"
    Yoshinobu Kawahara, Takehisa Yairi, and Kazuo Machida
    Seventh IEEE International Conference on Data Mining, 2007

    TODO:
    Implement the online recursive algorithm.

    We decided to do type hinting but not type checking as it requires too much boilerplate code. We recommend
    to input the specified types as the program will break in unforeseen ways otherwise.
    """

    def __init__(self, window_length: int, n_windows: int = None, lag: int = None, rank: int = 5, scale: bool = True,
                 method: str = 'offline', random_rank: int = None) -> None:
        """
        Initializing the Singular Spectrum Transformation (SST) requires setting a lot of parameters. See the parameters
        explanation for some intuition into the right choices. Currently, there are two SST methods from [1] and [2]
        available for use.

        :param window_length: This specifies the length of the time series (in samples), which will be used to extract
        the representative "eigensignals" to be compared before and after the lag. The windows length should be big
        enough to cover any wanted patterns (e.g. bigger than the periodicity of periodic signals).

        :param n_windows: This specifies the amount of consecutive time windows used to extract the "eigensignals" from
        the given time series. It should be big enough to cover different parts of the target behavior. If one does not
        specify this value, we use the rule of thumb and take as many time windows as you specified the length of the
        window (rule of thumb).

        :param lag: This parameter specifies the distance of the comparison for behaviors. In easy terms it tells the
        algorithms how far it should look into the future to find change in behavior to the current signal. If you do
        not specify this parameter, we use a rule of thumb and look ahead half of the window length you specified to
        cover the behavior.

        :param rank: This parameter specifies the amount of "eigensignals" which will be used to measure the
        dissimilarity of the signal in the future behavior. As a rule of thumb we take the five most dominant
        "eigensignals" if you do no specify otherwise.

        :param scale: Due to numeric stability we REALLY RECOMMEND scaling the signal into a restricted value range. Per
        default, we use a min max scaling to ensure a restricted value range. In the presence of extreme outliers this
        could cause problems, as the signal will be squished.

        :param method: Currently only the offline method of [1] is available.

        :param random_rank: In order to use the randomized singular value decomposition, one needs to provide a
        randomized rank (size of second matrix dimension for the randomized matrix) as specified in [3]. The lower
        this value, the faster the computation but the higher the error (as the approximation gets worse).
        """
        raise NotImplementedError('SI is not yet available.')
        # save the specified parameters into instance variables
        self.window_length = window_length
        self.n_windows = n_windows
        self.lag = lag
        self.rank = rank
        self.scale = scale
        self.method = method
        self.random_rank = random_rank

        # set some default values when they have not been specified
        if self.n_windows is None:
            self.n_windows = self.window_length
        if self.lag is None:
            # rule of thumb
            self.lag = max(self.n_windows//3, 1)
        if self.random_rank is None:
            # compute the rank as specified in [3] and
            # https://scikit-learn.org/stable/modules/generated/sklearn.utils.extmath.randomized_svd.html
            self.random_rank = self.rank + 10

        # specify the methods and their corresponding functions as lambda functions expecting only the
        # 1) future hankel matrix,
        # 2) the current hankel matrix and
        # 3) the feedback vector (e.g. for dominant eigenvector feedback)
        # all other parameter should be specified as values in the partial lambda function
        self.methods = {'offline': partial(_subspace_identification,
                                           rank=self.rank,
                                           randomized_rank=self.random_rank)
                        }

        # check whether the method is correct
        assert self.method in self.methods, f'Specified method {self.method} is not available in {self.methods.keys()}.'

    def transform(self, time_series: np.ndarray) -> np.ndarray:
        """
        This function calculate the anomaly score for each sample within the time series starting from an initial sample
        being the first sample of fitting the past hankel matrix (window_length + n_windows samples) and the new future
        hankel matrix (+lag). Values before that will be zero

        It also does some assertions regarding the specified parameters and whether they fit the time series.

        This function builds the interface to the jit compiled static function used to iterate over the array.

        :param time_series: 1D array containing the time series to be scored
        :return: anomaly score
        """

        # check the dimensions of the input array
        assert time_series.ndim == 1, "Time series needs to be an 1D array."

        # compute the starting point of the scoring (past and future hankel need to fit)
        starting_point = self.window_length + self.n_windows + self.lag
        assert starting_point < time_series.shape[0], "The time series is too short to score any points."

        # scale the time series (or just copy it if already scaled)
        if self.scale:
            time_series = normalization.min_max_scaling(time_series, min_val=1.0, max_val=2.0, inplace=False)
        else:
            time_series = time_series.copy()

        # get the changepoint scorer from the different methods
        scoring_function = self.methods[self.method]

        # start the scaling itself by calling the jit compiled staticmethod and return the result
        score = _transform(time_series=time_series, start_idx=starting_point, window_length=self.window_length,
                           n_windows=self.n_windows, lag=self.lag, scoring_function=scoring_function)
        return score


def _transform(time_series: np.ndarray, start_idx: int, window_length: int, n_windows: int, lag: int,
               scoring_function: Callable) -> np.ndarray:
    """
    Compute heavy and hopefully jit compilable score computation for the SST method. It does not do any parameter
    checking and can throw cryptic errors. It's only used for internal use as a private function.

    :param time_series: 1D array containing the time series to be scored
    :param start_idx: integer defining the start sample index for the score computation
    :param window_length: the size of the time series window for each column of the hankel matrix
    :param n_windows: amount of columns in the hankel matrix
    :param lag: sample distance between future and past hankel matrix
    """

    # initialize a scoring array with no values yet
    score = np.zeros_like(time_series)

    # iterate over all the values in the signal starting at start_idx computing the change point score
    for idx in range(start_idx, time_series.shape[0]):

        # compile the past hankel matrix (H1)
        hankel_past = lg.compile_hankel(time_series, idx-lag, window_length, n_windows)

        # compile the future hankel matrix (H2)
        hankel_future = lg.compile_hankel(time_series, idx, window_length, n_windows)

        # compute the score and save the returned feedback vector
        score[idx] = scoring_function(hankel_past, hankel_future)

    return score


def _subspace_identification(hankel_past: np.ndarray, hankel_future: np.ndarray, hanke_test,
                             rank: int, randomized_rank: int) -> (float, np.ndarray):
    """
    This class implements all the utility and functionality necessary to compute the subspace change point detection
    algorithm as described in (offline method)

    [1]
    "Change-Point Detection in Time-Series Data based on Subspace Identification"
    Yoshinobu Kawahara, Takehisa Yairi, and Kazuo Machida
    Seventh IEEE International Conference on Data Mining, 2007

    param hankel_past: the hankel matrix H1 before the change point
    :param hankel_future: the hankel matrix H2 after the change point
    :param rank: the amount of (approximated) eigenvectors as subspace of H1
    :param randomized_rank: the rank of the approximation used to construct the noise matrix
    :return: the change point score
    """

    # compute the covariance matrix for each time window and their cross covariance
    cov_pp = hankel_past.T @ hankel_past
    cov_ff = hankel_future.T @ hankel_future
    cov_fp = hankel_future.T @ hankel_past

    # compute the inverse square root of the matrices
    cov_pp = np.linalg.inv(sp.linalg.sqrtm(cov_pp))
    cov_ff = np.linalg.inv(sp.linalg.sqrtm(cov_ff))

    # compute the observability matrix of the system
    observability = cov_ff @ cov_fp @ cov_pp.T

    # compute the highest eigenvectors of the observability matrix
    _, eigenvectors = lg.randomized_singular_value_decomposition(observability, randomized_rank=randomized_rank)
    eigenvectors = eigenvectors[:, :rank] @ eigenvectors[:, :rank].T

    # compute the change point score
    return sum(vec.T @ vec - vec.T @ eigenvectors @ vec for vec in hankel_future.T)


if __name__ == '__main__':

    # make synthetic step function
    np.random.seed(123)
    length = 300
    x = np.hstack([1 * np.ones(length) + np.random.rand(length) * 1,
                   3 * np.ones(length) + np.random.rand(length) * 2,
                   5 * np.ones(length) + np.random.rand(length) * 1.5])
    x += np.random.rand(x.size)

    # create the sst method
    si = SubspaceIdentification(31)

    # make the scoring
    si.transform(x)


# this file contains the tests for the linalg.py file in the utils of this package
import pytest
import numpy as np
import changepoynt.utils.linalg as lg


class TestLinearAlgebra:
    def setup_method(self):
        # set a random seed
        np.random.seed(3455)

        # create a random square matrix
        self.A = np.random.rand(50, 50)

        # create random starting vector for the power method
        # and normalize it
        self.x0 = np.random.rand(50)
        self.x0 /= np.linalg.norm(self.x0)

        # create a tridiagonal matrix and the diagonal elements
        self.d = 5 * np.random.rand(100)
        self.e = -1 * np.random.rand(99)
        self.T = np.diag(self.d) + np.diag(self.e, k=1) + np.diag(self.e, k=-1)

        # compute the svd of the matrix A as we will need it for comparisons
        self.eigvecs, self.eigvals, _ = np.linalg.svd(self.A)

    def teardown_method(self):
        pass

    def test_power_method(self):
        eigval, eigvec = lg.power_method(self.A, self.x0, n_iterations=100)
        np.testing.assert_almost_equal(eigval, self.eigvals[0])
        np.testing.assert_almost_equal(np.abs(eigvec), np.abs(self.eigvecs[:, 0]))

    def test_eig_tridiag(self):
        # take a look at the highest half of the eigenvalues as lower one might be unstable
        # https://stackoverflow.com/questions/46345217/diagonalization-of-a-tridiagonal-symmetric-sparse-matrix-with-python
        amount = self.d.shape[0]//2
        tri_eigvals, tri_eigvecs = lg.tridiagonal_eigenvalues(self.d, self.e, amount)
        eigvecs, eigvals, _ = np.linalg.svd(self.T)
        np.testing.assert_almost_equal(tri_eigvals, eigvals[:amount])
        np.testing.assert_almost_equal(np.abs(tri_eigvecs), np.abs(eigvecs[:, :amount]))

    def test_rayleigh_ritz_svd(self):
        k = 20
        k_eigvals, k_eigvecs = lg.rayleigh_ritz_singular_value_decomposition(self.A, k)

        # order is not guaranteed by the method, so we need to sort in decreasing order
        idx = np.argsort(k_eigvals)[::-1]
        k_eigvals = k_eigvals[idx]
        k_eigvecs = k_eigvecs[:, idx]

        # compare the values
        np.testing.assert_almost_equal(k_eigvals, self.eigvals[:k])
        np.testing.assert_almost_equal(np.abs(k_eigvecs), np.abs(self.eigvecs[:, :k]))

    def test_randomized_svd(self):

        # set a threshold for the amount of eigenvectors we want to have and compute them
        k = 3
        k_eigvals, k_eigvecs = lg.randomized_singular_value_decomposition(self.A, 50)

        # compare the values
        np.testing.assert_almost_equal(k_eigvals[:k], self.eigvals[:k])
        np.testing.assert_almost_equal(np.abs(k_eigvecs[:, :k]), np.abs(self.eigvecs[:, :k]))

    def test_restarted_implicit_lanczos_svd(self):
        """
        !NOTE!
        This function currently only works for large eigenvalues. So we can mostly only compare the highest two.

        Is that a problem? We will see.
        :return:
        """

        # create a symmetric random matrix (as needed for this method)
        # as proposed in https://stackoverflow.com/questions/10806790/generating-symmetric-matrices-in-numpy
        a = np.random.rand(100, 100)
        a = np.tril(a) + np.tril(a, -1).T

        # test that the matrix is hermitian (symmetric for real valued matrices)
        np.testing.assert_almost_equal(a, a.T)

        # compute the svd of the matrix A as we will need it for comparisons
        eigvecs, eigvals, _ = np.linalg.svd(a)

        # get the results from the lanczos method
        k = 2
        k_eigvals, k_eigvecs = lg.implicit_restarted_lanczos_bidiagonalization(a, k, 2*k)

        # compare the values
        np.testing.assert_almost_equal(k_eigvals[::-1], eigvals[:k])
        np.testing.assert_almost_equal(np.abs(k_eigvecs)[:, ::-1], np.abs(eigvecs[:, :k]))


if __name__ == "__main__":
    pytest.main()



import numpy as np
from changepoynt.algorithms.base_algorithm import Algorithm
from changepoynt.algorithms.rulsif import RuLSIF


class ULSIF(Algorithm):
    """
    This class implements change point detection based on density ration estimation optimizing a least squares
    approach from:

    [1]
    "A least-squares approach to direct importance estimation"
    T. Kanamori, S. Hido, and M. Sugiyama.
    Journal of Machine Learning Research, 10:1391–1445, 2009.

    and can be implemented as a special version of with alpha = 0

    [2]
    Liu, Song, et al.
    "Change-point detection in time-series data by relative density-ratio estimation."
    Neural Networks 43 (2013): 72-83.

    which we are doing here.
    """

    def __init__(self, window_length: int = 10, n_windows: int = 50, lag: int = None, estimation_lag: int = None,
                 n_kernels: int = 100) -> None:
        """
        This defines all necessary parameters for the RuLSIF to work.
        :param window_length: the length of the windows we want to compare the densities for (k in the paper)
        :param n_windows: the amount of windows (n in the paper)
        :param lag: the difference in time between the past and the future comparison window
        :param estimation_lag: how often we should re-estimate lambda and sigma for the gaussian kernels
        :param n_kernels: the amount of gaussian kernels for the density ratio
        """
        self.window_length = window_length
        self.n_windows = n_windows
        self.lag = lag
        self.estimation_lag = estimation_lag
        self.n_kernels = n_kernels

        # create the specialized version of RuLSIF
        self.transformer = RuLSIF(window_length=self.window_length, n_windows=self.n_windows, lag=self.lag,
                                  estimation_lag=self.estimation_lag, n_kernels=self.n_kernels, alpha=0.0)

    def transform(self, time_series: np.ndarray):
        return self.transformer.transform(time_series)


def short_test():
    from time import time
    # make synthetic step function
    np.random.seed(123)
    length = 300
    x = np.hstack([1 * np.ones(length) + np.random.rand(length) * 1,
                   3 * np.ones(length) + np.random.rand(length) * 2,
                   5 * np.ones(length) + np.random.rand(length) * 1.5])
    x += np.random.rand(x.size)

    # create the rulsif method
    ruli = ULSIF()

    # make the scoring
    start = time()
    score = ruli.transform(x)
    print((time() - start) / (length * 3))
    print(score[600:610])


if __name__ == '__main__':
    short_test()



"""
This file contains the interface for the base algorithm used in the package.
"""
from abc import ABC, abstractmethod
import numpy as np


class Algorithm(ABC):

    @abstractmethod
    def transform(self, time_series: np.ndarray):
        raise NotImplementedError



import numpy as np
from scipy import linalg
from abc import ABC, abstractmethod

"""
Taken an modified from:
https://github.com/DeepOmics/rulsif_abrupt-change_detection
"""


class Estimator(ABC):
    """
    Marker interface for all Estimators. Provides basic apply() style
    interface and a show() method for debugging
    """
    @abstractmethod
    def apply(self, x, y):
        raise NotImplementedError

    @abstractmethod
    def train(self, x, y):
        raise NotImplementedError

    @abstractmethod
    def show(self):
        raise NotImplementedError


class GaussianKernel:

    def __init__(self, sigma=1.0):
        self.sigmaWidth = sigma

    def computeDistance(self, samples=None, sampleMeans=None):
        """
        Compute the distances between points in the sample's feature space
        to points along the center of the distribution
        """
        (sampleRows, sampleCols) = samples.shape
        (meanRows, meanCols) = sampleMeans.shape

        squaredSamples = np.sum(samples**2, 0)
        squaredMeans = np.sum(sampleMeans**2, 0)

        return np.tile(squaredMeans, (sampleCols, 1)) + np.tile(squaredSamples[:, None], (1, meanCols))\
            - 2 * np.dot(samples.T, sampleMeans)

    def apply(self, samples=None, sampleMeans=None):
        """
        Computes an n-dimensional Gaussian/RBF kernel matrix by taking points
        in the sample's feature space and maps them to kernel coordinates in
        Hilbert space by calculating the distance to each point in the sample
        space and taking the Gaussian function of the distances.
           K(X,Y) = exp( -(|| X - Y ||^2) / (2 * sigma^2) )
        where X is the matrix of data points in the sample space,
              Y is the matrix of gaussian centers in the sample space
             sigma is the width of the gaussian function being used
        """
        squaredDistance = self.computeDistance(samples, sampleMeans)
        return np.exp(-squaredDistance / (2 * (self.sigmaWidth**2)))


class AlphaRelativeDensityRatioEstimator(Estimator):
    """
    Computes the alpha-relative density ratio estimate of P(X_ref) and P(X_test)
    The alpha-relative density ratio estimator, r_alpha(X), is given by the
    following kernel model:
    g(X; theta) = SUM( (theta_l * K(X, X_centers_l)), l=0, n )
    where theta is a vector of parameters [theta_1, theta_2, ..., theta_l]^T
    to be learned from the data samples. The parameters theta in the model
    g(X; theta) is calculated by solving the following optimization problem:
      theta_hat = argmin [ ( (1/2) * theta^T * H_hat * theta) - (h_hat^T * theta) + ( lambda/2 * theta^T * theta) ]
    where the expression (lamba/2 * theta^T * theta), with lambda >= 0, is
    a regularization term used to penalize against overfitting
    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto Yamada,
    Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama. NIPS,
    page 594-602. (2011)
    """
    alphaConstraint = None
    sigmaWidth = None
    lambdaRegularizer = None
    kernelBasis = None

    def __init__(self, alphaConstraint=0.0, sigmaWidth=1.0, lambdaRegularizer=0.0, kernelBasis=1):
        self.alphaConstraint = alphaConstraint
        self.sigmaWidth = sigmaWidth
        self.lambdaRegularizer = lambdaRegularizer
        self.kernelBasis = kernelBasis

    def train(self, x, y):
        pass

    def show(self):
        pass

    def apply(self, referenceSamples=None, testSamples=None, gaussianCenters=None):
        """
        Computes the alpha-relative density ratio, r_alpha(X), of P(X_ref) and P(X_test)
          r_alpha(X) = P(Xref) / (alpha * P(Xref) + (1 - alpha) * P(X_test)
        Returns density ratio estimate at X_ref, r_alpha_ref, and at X_test, r_alpha_test
        """
        # Apply the kernel function to the reference and test samples
        K_ref = GaussianKernel(self.sigmaWidth).apply(referenceSamples, gaussianCenters).T
        K_test = GaussianKernel(self.sigmaWidth).apply(testSamples, gaussianCenters).T

        # Compute the parameters, theta_hat, of the density ratio estimator
        H_hat = AlphaRelativeDensityRatioEstimator.H_hat(self.alphaConstraint, K_ref, K_test)
        h_hat = AlphaRelativeDensityRatioEstimator.h_hat(K_ref)
        theta_hat = AlphaRelativeDensityRatioEstimator.theta_hat(H_hat, h_hat, self.lambdaRegularizer, self.kernelBasis)

        # Estimate the density ratio, r_alpha_ref = r_alpha(X_ref)
        r_alpha_ref = AlphaRelativeDensityRatioEstimator.g_of_X_theta(K_ref, theta_hat).T
        # Estimate the density ratio, r_alpha_test = r_alpha(X_test)
        r_alpha_test = AlphaRelativeDensityRatioEstimator.g_of_X_theta(K_test, theta_hat).T

        # print(r_alpha_ref)
        # print(r_alpha_test)

        # distance_type = ['Hellinger Distance',]

        # HD  = abs(r_alpha_ref- r_alpha_test)

        return r_alpha_ref, r_alpha_test

    @staticmethod
    def H_hat(alpha=0.0, KernelMatrix_refSamples=None, KernelMatrix_testSamples=None):
        """
        Calculates the H_hat term of the theta_hat optimization problem
        """
        N_ref = KernelMatrix_refSamples.shape[1]
        N_test = KernelMatrix_testSamples.shape[1]

        H_hat = (alpha / N_ref) * np.dot(KernelMatrix_refSamples, KernelMatrix_refSamples.T) + \
                ((1.0 - alpha) / N_test) * np.dot(KernelMatrix_testSamples, KernelMatrix_testSamples.T)

        return H_hat

    @staticmethod
    def h_hat(KernelMatrix_refSamples):
        """
        Calculates the h_hat term of the theta_hat optimization problem
        """
        h_hat = np.mean(KernelMatrix_refSamples, 1)

        return h_hat

    @staticmethod
    def theta_hat(H_hat=None, h_hat=None, lambdaRegularizer=0.0, kernelBasis=None):
        """
        Calculates theta_hat given H_hat, h_hat, lambda, and the kernel basis function
        Treat as a system of linear equations and find the exact, optimal solution
        """
        theta_hat = linalg.solve(H_hat + (lambdaRegularizer * np.eye(kernelBasis)), h_hat)

        return theta_hat

    @staticmethod
    def J_of_theta(alpha=0.0, g_Xref_theta=None, g_Xtest_theta=None):
        """
        Calculates the squared error criterion, J
        """
        return ((alpha / 2.0) * (np.mean(g_Xref_theta ** 2)) +
                ((1 - alpha) / 2.0) * (np.mean(g_Xtest_theta ** 2)) -
                np.mean(g_Xref_theta))

    @staticmethod
    def g_of_X_theta(KernelMatrix_samples=None, theta_hat=None):
        """
        Calculate the alpha-relative density ratio kernel model
        """
        return np.dot(KernelMatrix_samples.T, theta_hat)


class PearsonRelativeDivergenceEstimator(Estimator):
    """
    Calculates the alpha-relative Pearson divergence score
    The alpha-relative Pearson divergence is given by the following expression:
      PE_alpha = -(alpha/2(n_ref)) * SUM(r_alpha(X_ref_i)^2, i=0, n_ref)        -
                  ((1-alpha)/2(n_test)) * SUM(r_alpha(X_test_j)^2, j=0, n_test) +
                  (1/n_ref) * SUM(r_alpha(X_ref_i), i=0, n_ref)                 -
                  1/2
    where r_alpha(X) is the alpha-relative density ratio estimator and is given by
    the following kernel model:
      g(X; theta) = SUM( (theta_l * K(X, X_centers_l)), l=0, n )
    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto
    Yamada, Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama.
    NIPS, page 594-602. (2011)
    """
    alphaConstraint = None
    sigmaWidth = None
    lambdaRegularizer = None
    kernelBasis = None

    def __init__(self, alphaConstraint=0.0, sigmaWidth=1.0, lambdaRegularizer=0.0, kernelBasis=1):
        self.alphaConstraint = alphaConstraint
        self.sigmaWidth = sigmaWidth
        self.lambdaRegularizer = lambdaRegularizer
        self.kernelBasis = kernelBasis

    def train(self, x, y):
        pass

    def show(self):
        pass

    def apply(self, referenceSamples=None, testSamples=None, gaussianCenters=None):
        """
        Calculates the alpha-relative Pearson divergence score
        """
        densityRatioEstimator = AlphaRelativeDensityRatioEstimator(self.alphaConstraint,
                                                                   self.sigmaWidth,
                                                                   self.lambdaRegularizer,
                                                                   self.kernelBasis)

        # Estimate alpha relative density ratio and pearson divergence score
        (r_alpha_Xref, r_alpha_Xtest) = densityRatioEstimator.apply(referenceSamples, testSamples, gaussianCenters)

        PE_divergence = (np.mean(r_alpha_Xref) -
                         (0.5 * (self.alphaConstraint * np.mean(r_alpha_Xref ** 2) +
                                 (1.0 - self.alphaConstraint) * np.mean(r_alpha_Xtest ** 2))) - 0.5)

        return PE_divergence, r_alpha_Xtest


class RULSIF(Estimator):
    """
    Estimates the alpha-relative Pearson Divergence via Least Squares Relative
    Density Ratio Approximation
    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto
    Yamada, Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama.
    NIPS, page 594-602. (2011)
    """

    def __init__(self, alpha: float = 0.0, sigma: float = None, lambda_regularization: float = None,
                 n_kernels: int = 100, cross_folds: int = 5, verbose: bool = False):

        self.alphaConstraint = alpha
        self.sigmaWidth = sigma
        self.lambdaRegularizer = lambda_regularization
        self.kernelBasis = n_kernels
        self.crossFolds = cross_folds
        self.verbose = verbose

    def getMedianDistanceBetweenSamples(self, sampleSet=None):
        """
        Jaakkola's heuristic method for setting the width parameter of the Gaussian
        radial basis function kernel is to pick a quantile (usually the median) of
        the distribution of Euclidean distances between points having different
        labels.
        Reference:
        Jaakkola T S, Haussler D. Exploiting Generative Models in Discriminative Classifiers[J].
        Advances in Neural Information Processing Systems, 1998, 11(11):487--493.
        Jaakkola, M. Diekhaus, and D. Haussler. Using the Fisher kernel method to detect
        remote protein homologies. In T. Lengauer, R. Schneider, P. Bork, D. Brutlad, J.
        Glasgow, H.- W. Mewes, and R. Zimmer, editors, Proceedings of the Seventh
        International Conference on Intelligent Systems for Molecular Biology.
        """
        numrows = sampleSet.shape[0]
        samples = sampleSet

        G = np.sum((samples * samples), axis=1)
        Q = np.tile(G[:, None], (1, numrows))
        R = np.tile(G, (numrows, 1))

        distances = Q + R - 2 * np.dot(samples, samples.T)

        distances = distances - np.tril(distances)
        distances = distances.reshape(numrows ** 2, 1, order="F").copy()

        return np.sqrt(0.5 * np.median(distances[distances > 0]))

    def computeGaussianWidthCandidates(self, referenceSamples=None, testSamples=None):
        """
        Compute a candidate list of Gaussian kernel widths. The best width will be
        selected via cross-validation
        """
        allSamples = np.c_[referenceSamples, testSamples]
        medianDistance = self.getMedianDistanceBetweenSamples(allSamples.T)

        return medianDistance * np.array([0.6, 0.8, 1, 1.2, 1.4])

    def generateRegularizationParams(self):
        """
        Generatees a candidate list of regularization parameters to be used
        with the L1 regularizer term of RULSIF optimization problem.  The
        best regularizer parameter will be chosen via cross-validation
        """
        return 10.0 ** np.array([-3, -2, -1, 0, 1])

    def generateAllGaussianCenters(self, referenceSamples=None):
        """
        Generates kernels in the region where the P(X_reference) takes large values
        """
        self.kernelBasis = referenceSamples.shape[1]
        return referenceSamples[:, np.r_[0:self.kernelBasis]]

    def generateRandomGaussianCenters(self, referenceSamples=None):
        """
        Randomly chooses Gaussian centers as an optimization
        """
        numcols = referenceSamples.shape[1]
        referenceSampleIdxs = np.random.permutation(numcols)

        self.kernelBasis = min(self.kernelBasis, numcols)
        return referenceSamples[:, referenceSampleIdxs[0:self.kernelBasis]]

    def generateFirstNGaussianCenters(self, referenceSamples=None):
        """
        Chooses the firts N samples as Gaussian centers as an optimization
        """
        numcols = referenceSamples.shape[1]
        self.kernelBasis = min(self.kernelBasis, numcols)
        return referenceSamples[:, np.r_[0:self.kernelBasis]]

    def generateGaussianCenters(self, referenceSamples=None):
        """
        Choose Gaussian centers based on a strategy
        """
        gaussianCenters = self.generateAllGaussianCenters(referenceSamples)

        return gaussianCenters

    def computeModelParameters(self, referenceSamples=None, testSamples=None, gaussianCenters=None):
        """
        Computes model parameters via k-fold cross validation process
        """
        (refRows, refCols) = referenceSamples.shape
        (testRows, testCols) = testSamples.shape

        sigmaWidths = self.computeGaussianWidthCandidates(referenceSamples, testSamples)
        lambdaCandidates = self.generateRegularizationParams()

        # Initialize cross validation scoring matrix
        crossValidationScores = np.zeros((np.size(sigmaWidths), np.size(lambdaCandidates)))

        # Initialize a cross validation index assignment list
        referenceSamplesCVIdxs = np.random.permutation(refCols)
        referenceSamplesCVSplit = np.floor(np.r_[0:refCols] * self.crossFolds / refCols)
        testSamplesCVIdxs = np.random.permutation(testCols)
        testSamplesCVSplit = np.floor(np.r_[0:testCols] * self.crossFolds / testCols)

        # Initiate k-fold cross-validation procedure. Using variable
        # notation similar to the RULSIF formulas.
        for sigmaIdx in np.r_[0:np.size(sigmaWidths)]:

            # (re-)Calculate the kernel matrix using the candidate sigma width
            sigma = sigmaWidths[sigmaIdx]
            K_ref = GaussianKernel(sigma).apply(referenceSamples, gaussianCenters).T
            K_test = GaussianKernel(sigma).apply(testSamples, gaussianCenters).T

            # Initialize a new result matrix for the current sigma candidate
            foldResult = np.zeros((self.crossFolds, np.size(lambdaCandidates)))

            for foldIdx in np.r_[0:self.crossFolds]:

                K_ref_trainingSet = K_ref[:, referenceSamplesCVIdxs[referenceSamplesCVSplit != foldIdx]]
                K_test_trainingSet = K_test[:, testSamplesCVIdxs[testSamplesCVSplit != foldIdx]]
                H_h_KthFold = AlphaRelativeDensityRatioEstimator.H_hat(self.alphaConstraint, K_ref_trainingSet,
                                                                       K_test_trainingSet)
                h_h_KthFold = AlphaRelativeDensityRatioEstimator.h_hat(K_ref_trainingSet)

                for lambdaIdx in np.r_[0:np.size(lambdaCandidates)]:
                    lambdaCandidate = lambdaCandidates[lambdaIdx]

                    theta_h_KthFold = AlphaRelativeDensityRatioEstimator.theta_hat(H_h_KthFold, h_h_KthFold,
                                                                                   lambdaCandidate, self.kernelBasis)

                    # Select the subset of the kernel matrix not used in the training set
                    # for use as the test set to validate against
                    K_ref_testSet = K_ref[:, referenceSamplesCVIdxs[referenceSamplesCVSplit == foldIdx]]
                    K_test_testSet = K_test[:, testSamplesCVIdxs[testSamplesCVSplit == foldIdx]]

                    r_alpha_Xref = AlphaRelativeDensityRatioEstimator.g_of_X_theta(K_ref_testSet, theta_h_KthFold)
                    r_alpha_Xtest = AlphaRelativeDensityRatioEstimator.g_of_X_theta(K_test_testSet, theta_h_KthFold)

                    # Calculate the objective function J(theta) under the current parameters
                    J = AlphaRelativeDensityRatioEstimator.J_of_theta(self.alphaConstraint, r_alpha_Xref, r_alpha_Xtest)

                    foldResult[foldIdx, lambdaIdx] = J

                crossValidationScores[sigmaIdx, :] = np.mean(foldResult, 0)

        crossValidationMinScores = np.min(crossValidationScores, axis=1)
        crossValidationMinIdxForLambda = crossValidationScores.argmin(1)
        crossValidationMinIdxForSigma = crossValidationMinScores.argmin()

        optimalSigma = sigmaWidths[crossValidationMinIdxForSigma]
        optimalLambda = lambdaCandidates[crossValidationMinIdxForLambda[crossValidationMinIdxForSigma]]

        return (optimalSigma, optimalLambda)

    def train(self, referenceSamples=None, testSamples=None):
        """
        Learn the proper model parameters
        """

        self.gaussianCenters = self.generateGaussianCenters(referenceSamples)

        (optimalSigma, optimalLambda) = self.computeModelParameters(referenceSamples, testSamples, self.gaussianCenters)

        self.sigmaWidth = optimalSigma
        self.lambdaRegularizer = optimalLambda

    def apply(self, referenceSamples=None, testSamples=None):
        """
        Estimates the alpha-relative Pearson divergence as determined by the relative
        ratio of probability densities:
           P(ReferenceSamples[x]) / (alpha * P(ReferenceSamples[x]) + (1 - alpha) * P(TestSamples[x]))
        from samples:
           ReferenceSamples[x_i] | ReferenceSamples[x_i] in R^{d}, with i=1 to ReferenceSamples{N}
        drawn independently from P(ReferenceSamples[x])
        and from samples:
           TestSamples[x_j] | TestSamples[x_j] in R^{d}, with j=1 to TestSamples{N}
        drawn independently from P(TestSamples[x])
        After the model hyper-parameters have been learned and chosen by the train()
        method, the RULSIF algorithm can be applied repeatedly on both in-sample and out
        of sample data
        """

        if self.gaussianCenters is None or self.kernelBasis is None:
            raise Exception("Missing kernel basis function parameters")

        if self.sigmaWidth == 0.0 or self.lambdaRegularizer == 0.0:
            raise Exception("Missing model selection parameters")

        divergenceEstimator = PearsonRelativeDivergenceEstimator(self.alphaConstraint, self.sigmaWidth,
                                                                 self.lambdaRegularizer, self.kernelBasis)
        (PE_alpha, r_alpha_Xtest) = divergenceEstimator.apply(referenceSamples, testSamples, self.gaussianCenters)

        self.show("RULSIF Results", self.sigmaWidth, self.lambdaRegularizer, PE_alpha, self.verbose)

        return PE_alpha

    def show(self, displayName=None, optimalSigma=0.0, optimalLambda=0.0, PE_alpha=0.0, verbose=False):
        if not verbose:
            return

        print("[" + displayName + "]\n")
        print("Alpha Constraint         : " + str(self.alphaConstraint))
        print("Kernel Basis Functions   : " + str(self.kernelBasis))
        print("Basis Function Width     : " + str(optimalSigma))
        print("Regularization Parameter : " + str(optimalLambda))
        print("Pearson Divergence Score : " + str(PE_alpha))
        print("\n")
        print("---------------")